/*
 * Secure FileCheck
 *	by shad0w and Vision
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>

#define BUILD 0
#define COMPARE 1

void cmp(char* buf, struct stat* stats, char* file);
void report_err(char* file, char* err);
void usage(char* name);
void parse(char* argv[], int argc, int& mode);
char* get_buf(char* rfile, int mode);
void build_db(char* dbfile, char* lbuf);
int get_nfiles(char* buf);
void get_stats(struct stat* stats[], char* files[], int nfiles);
void get_dstrings(char* databuf, char* dstrings[], char* files[]);

int main(int argc, char** argv)
{
	int mode;
	
	parse(argv, argc, mode);
	if (mode == BUILD) {
		char* listbuf;
		
		listbuf = get_buf(argv[2], mode);
		build_db(argv[3], listbuf);
		free(listbuf);
	}
	else if (mode == COMPARE) {
		char* databuf;
		int nfiles;
		char (*files[nfiles]);
		char (*dstrings[nfiles]);
		struct stat (*stats[nfiles]);

		databuf = get_buf(argv[2], mode);
		nfiles = get_nfiles(databuf);
		get_dstrings(databuf, dstrings, files);
		free(databuf);
		get_stats(stats, files, nfiles);
		for (int i = 0; i < nfiles; i++) {
			cmp(dstrings[i], stats[i], files[i]);
			free(stats[i]);
		}
	}
	exit(0);
}

/*
 * Turn the databuf into an array of strings, and setup an array
 * of filenames from the databuf.
 */
void get_dstrings(char* databuf, char* dstrings[], char* files[])
{
	int len;
	int count;
	
	count = 0;
	len = strlen(databuf);
	for (int i = 0; i < len; i++)
		if (databuf[i] == '\n') {
			dstrings[count] = (char*)calloc(1, i + 1);
			strncpy(dstrings[count], databuf, i);
			memmove(databuf, &databuf[i + 1], len - i - 1);
			memset(&databuf[len - i - 1], '\0', i + 1);
			len -= i + 1;
			i = 0;
			count++;
		}
	for (int i = 0; i < count; i++) {
		len = strlen(dstrings[i]);
		for (int j = 0; j < len; j++)
			if (dstrings[i][j] == ':') {
				files[i] = (char*)calloc(1, j + 1);
				strncpy(files[i], dstrings[i], j);
				break;
			}
	}
}

/*
 * A quick routine to get an array of file stats
 * Because program is expected to run quietly there
 * is no whining about errors here.
 */
void get_stats(struct stat* stats[], char* files[], int nfiles)
{
	for (int i = 0; i < nfiles; i++) {
		stats[i] = (struct stat*)malloc(sizeof(struct stat));
		stat(files[i], stats[i]);
	}
	return;
}

/*
 * This was the neatest way to run things...
 */
int get_nfiles(char* buf)
{
	int nfiles;
	int len;
	
	nfiles = 0;
	len = strlen(buf);
	for (int i = 0; i < len; i++)
		if (buf[i] == '\n')
			nfiles++;
	return nfiles;
}

/* 
 * Reads the file list and gets size and mod-time info on the files.
 * Stores the results in the specified file.
 */
void build_db(char* dbfile, char* lbuf)
{
	struct stat stats;
	int nfiles;
	int len;
	FILE* file;
	
	umask(077);
	len = strlen(lbuf);
	file = fopen(dbfile, "w");
	if (file == (FILE*)NULL) {
		perror("fopen");
		free(lbuf);
		exit(1);
	}
	nfiles = get_nfiles(lbuf);
	printf("Storing stats for:  \n");
	for (int i = 0; i < nfiles; i++) {
		for (int j = 0; j < len; j++) {
			if (lbuf[j] == '\n')
				lbuf[j] = '\0';
			else continue;
			printf("\t%s.....\n", lbuf);
			if (stat(lbuf, &stats) == -1)
				if (errno = ENOENT) {
					if (strlen(lbuf) > 0) {
					   printf("-*- Line %i -*-\t", i + 1);
					   printf("File not found:  %s\n", lbuf);
					}
					memmove(lbuf, &lbuf[j + 1], \
					           len - j - 1);
					break;
				}
				else {
					free(lbuf);
					perror("stat");
					exit(1);
				}
			fprintf(file, "%s:%i:", lbuf, stats.st_size);
			fprintf(file, "%i\n", stats.st_mtime);
			memmove(lbuf, &lbuf[j + 1], len - j - 1);
			break;
		}
	}
	if (fclose(file) == EOF) {
		perror("fclose");
		free(lbuf);
		exit(1);
	}
	printf("-*- Database  %s  has been updated -*-\n", dbfile);
	return;	
}

/* 
 * Code to read a file into a buffer.
 */
char* get_buf(char* rfile, int mode)
{
	char* buf;
	struct stat stats;
	int file;

	if (stat(rfile, &stats) == -1) {
		perror("stat");
		exit(1);
	}
	buf = (char*)calloc(1, stats.st_size + 1);
	if (buf == (char*)NULL) {
		printf("Couldn't get memory\n");
		exit(1);
	}
	file = open(rfile, O_RDONLY);
	if (file == -1) {
		perror("open");
		free(buf);
		exit(1);
	}
	read(file, buf, stats.st_size);
	close(file);
	if (mode = BUILD)
		buf[stats.st_size] = '\n';
	return buf;
}

/*
 * Emails errors to root@localhost.
 */
void report_err(char* file, char* err)
{
	int s;
	struct sockaddr_in sin;
	
	sin.sin_family = AF_INET;
	sin.sin_port = htons(25);
	sin.sin_addr.s_addr = inet_addr("127.0.0.1");
	memset(sin.sin_zero, 0, 8);
	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s == -1)
		return;
	if (connect(s, (struct sockaddr*)&sin, sizeof(struct sockaddr)) == -1)
		return;
	sleep(1);
	send(s, "helo localhost\n", 15, 0);
	sleep(1);
	send(s, "mail from:NCSfck@localhost\n", 27, 0);
	sleep(1);
	send(s, "rcpt to:root@localhost\n", 23, 0);
	sleep(1);
	send(s, "data\n", 5, 0);
	sleep(1);
	send(s, "NCS File Check reported the following error:\n\t", 46, 0);
	sleep(1);
	send(s, file, strlen(file), 0);
	sleep(1)	;
	send(s, " : ", 3, 0);
	sleep(1);
	send(s, err, strlen(err), 0);
	sleep(1);
	send(s, "\n.\n", 3, 0);
	sleep(1);
	shutdown(s, 2);
	sleep(1);
	return;
}

/*
 * Blah...
 */
void usage(char* name)
{
	printf("(r0ot) Secure FileCheck by shad0w and Vision\nUsage:\n");
	printf("To build a database of file stats:\n\t");
	printf("%s -b <file list> <database file>\n", name);
	printf("To compare current file stats to a database,\n");
	printf("this should be run regularly, like from cron:\n\t");
	printf("%s -c <database file>\n", name);
	exit(0);
}

/* 
 * Compare current file stats to those in the database.
 * Any mismatches will result in a report_err() call.
 */
void cmp(char* buf, struct stat* stats, char* file)
{
	int len;

	len = strlen(buf);
	for (int i = 0; i < len; i++)
		if (buf[i] == ':') {
			memmove(buf, &buf[i + 1], len - i - 1);
			memset(&buf[len - i - 1], '\0', i + 1);
			len -= i + 1;
			break;
		}
	for (int i = 0; i < len; i++)
		if (buf[i] == ':') {
			buf[i] = '\0';
			if (stats->st_size != atoi(buf))
				report_err(file, "Filesize has changed");
			memmove(buf, &buf[i + 1], len - i - 1);
			memset(&buf[len - i - 1], '\0', i + 1);
			len -= i + 1;
			break;
		}
	if (stats->st_mtime != atoi(buf))
		report_err(file, "File has been modified");
}

/*
 * I never did figure out getopt()...
 */
void parse(char* argv[], int argc, int& mode)
{
	if (argc < 3)
		usage(argv[0]);
	if (strcmp(argv[1], "-b") != 0 && strcmp(argv[1], "-c"))
		usage(argv[0]);
	if (!strcmp(argv[1], "-b") && argc != 4)
		usage(argv[0]);
	else if (!strcmp(argv[1], "-b") && argc == 4)
		mode = BUILD;
	if (!strcmp(argv[1], "-c") && argc != 3)
		usage(argv[0]);
	else if (!strcmp(argv[1], "-c") && argc == 3)
		mode = COMPARE;
}
